# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""ORM mixin providing orm-like functionality for an index.

Works with dynamic and static mappings.
"""
from abc import ABCMeta
from collections import defaultdict

from . import fields
from .exceptions import FieldConflictError, MissingMappingError


class OrmMixin(object):
    """Provides dot-notation for accessing index field names.

    Attributes:
        fields: Fields that exist within an index at a root level
        namespaces: Namespaces that exist within an index at a root level
        dtypes: Dtypes of all fields (including namespaced)
    """

    __metaclass__ = ABCMeta
    _type_mapping = {
        'integer': fields.Integer,
        'float': fields.Float,
        'scaled_float': fields.Decimal,
        'keyword': fields.String,
        'text': fields.String,
        'date': fields.Date,
        'boolean': fields.Boolean
    }

    @property
    def fields(self):
        """Fields that exist within an index at a root level."""
        return [i._name
                for i in vars(self).values()
                if isinstance(i, fields.Field)]

    @property
    def namespaces(self):
        """Namespaces that exist within an index at a root level."""
        return [i._name
                for i in vars(self).values()
                if isinstance(i, fields.Namespace)]

    @property
    def dtypes(self):
        """Dtypes of all fields (including namespaced)."""
        return self.__dtypes(self)

    @staticmethod
    def __dtypes(obj):
        """Recursively loads dtypes from nested namespaces."""
        dtypes = defaultdict(dict)
        for k, v in vars(obj).items():
            if k == 'parent':
                continue
            if isinstance(v, fields.Field):
                dtypes[v._name] = v.dtype
            elif isinstance(v, fields.Namespace):
                dtypes[v._name].update(OrmMixin.__dtypes(v))
        return dict(dtypes)

    def get_namespace(self, key):
        """Return the namespace.

        Equivalent to `ds.<namespace>` in dot notation

        Args:
            key (str): The name of the namespace

        Returns:
            field.Namespace
        """
        return getattr(self, key)

    def _load_orm(self):
        """Map elasticsearch index fields to attributes."""
        raw = self._es.indices.get_mapping(index=self.index,
                                           include_type_name=False)
        _, raw = raw.popitem()
        properties = raw['mappings'].get('properties')
        if properties:
            self.__parse_properties(properties)
            return properties
        raise MissingMappingError(self.index)

    def __parse_properties(self, properties, namespace=None):
        """Recursively create obj attributes from index properties."""
        for name, definition in properties.items():
            if 'type' in definition:
                self.__add_field(name, definition['type'], namespace)
            if 'properties' in definition:
                ns = self.__create_namespace(name, namespace)
                self.__parse_properties(definition['properties'], ns)

    def __create_namespace(self, name, namespace):
        """Create a namespace and adds it as an object attribute."""
        if namespace is not None:
            ns = fields.Namespace(name, namespace)
            setattr(namespace, name, ns)
        elif hasattr(self, name):
            raise FieldConflictError(name)
        else:
            ns = fields.Namespace(name, self)
            setattr(self, name, ns)
        return ns

    def __add_field(self, name, dtype, namespace):
        """Create a field and adds it as an object attribute."""
        Field = self._type_mapping.get(dtype, fields.Dummy)
        if namespace is not None:
            f = Field(name, namespace)
            setattr(namespace, name, f)
        elif hasattr(self, name):
            raise FieldConflictError(name)
        else:
            f = Field(name, self)
            setattr(self, name, f)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""Module for managing elasticsearch / bamboo configuration settings."""
import os
from collections import MutableMapping

from elasticsearch import Elasticsearch

from .utils import dict_to_params


def set_connection(func):
    """Decorate function to recreate elasticsearch connection object."""
    def wrapper(obj, *args, **kwargs):
        func(obj, *args, **kwargs)
        obj.connection = Elasticsearch(**obj)
    return wrapper


class Config(MutableMapping):
    """Configuration storage class."""

    @set_connection
    def __init__(self, **kwargs):
        """Init Config."""
        self.config_files = ('setup.cfg', 'tox.ini', '.bamboo')
        self.connection = None
        self._config = {}
        self.__search(kwargs)

    def __repr__(self):
        return '{}({})'.format(
            type(self).__name__,
            dict_to_params(self._config)
        )

    @set_connection
    def __call__(self, **kwargs):
        """Set connection arguments."""
        self._config.update(kwargs)

    def __getstate__(self):
        return self._config

    @set_connection
    def __setstate__(self, state):
        self._config = state

    def __iter__(self):
        return iter(self._config)

    def __len__(self):
        return len(self._config)

    @set_connection
    def __setitem__(self, key, value):
        self._config[key] = value

    def __getitem__(self, key):
        return self._config.get(key)

    @set_connection
    def __delitem__(self, key):
        del self._config[key]

    def __search(self, kwargs):
        self.__from_env()
        self.__from_files()
        self.update(kwargs)

    def __from_env(self):
        for k, v in os.environ.items():
            lower_k = k.lower()
            if lower_k.startswith('bamboo'):
                var_idx = lower_k.index('_') + 1
                # convert to array if spaces in env var string
                v = v.split(' ') if ' ' in v else v
                self[lower_k[var_idx:]] = v

    def __from_files(self):
        for fname in self.config_files:
            pass


config = Config()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""Field and namespace objects representing Elasticsearch field types."""
import warnings
from abc import ABCMeta
from copy import deepcopy
from datetime import datetime, timedelta
from functools import wraps

from . import queries


def check_inversion(func):
    """Decorate a method for invertible operations."""
    @wraps(func)
    def wrapper(obj, *args, **kwargs):
        condition = func(obj, *args, **kwargs)
        if obj._inverted is True:
            condition = queries.Bool(must_not=condition)
        return condition
    return wrapper


class Base(object):
    """Parent class for field and namespace classes.

    Attributes:
        name: Full field name including any parents
        parent: Parent for a field
    """

    __metaclass__ = ABCMeta

    def __init__(self, name, parent):
        """Init Base.

        Args:
            name (str): The name for the class
        """
        self.parent = parent
        self._name = name
        self._inverted = False

    def __invert__(self):
        new = deepcopy(self)
        new._inverted = True
        return new

    def __repr__(self):
        return '{}({})'.format(type(self).__name__, self._name)

    @property
    def name(self):
        """Full field name including any parents."""
        if isinstance(self.parent, Base):
            return self.parent.name + '.' + self._name
        return self._name

    @property
    def root(self):
        """Dataframe where a field is attached."""
        return self.__get_root(self)

    @staticmethod
    def __get_root(obj):
        """Recurse parent objects until root dataframe."""
        parent = obj.parent
        if not isinstance(parent, Base):
            return parent
        return Base.__get_root(parent)

    @check_inversion
    def exists(self):
        """Condition checking whether the field exists."""
        return queries.Exists(self.name)


class Namespace(Base):
    """Namespace field meant to hold children fields.

    Attributes:
        name: Full field name including any parents
        parent: Parent for a field
        fields: Children fields held by namespace
    """

    def __getitem__(self, key):
        return getattr(self, key)

    @property
    def fields(self):
        """Fields that exist within a namespace."""
        return [i._name
                for i in vars(self).values()
                if isinstance(i, Field)]


class Field(Base):
    """Base field class for different dtypes.

    Attributes:
        name: Full field name including any parents
        parent: Parent for a field
    """

    __metaclass__ = ABCMeta

    @check_inversion
    def __eq__(self, value):
        return queries.Term(self.name, value)

    def __ne__(self, value):
        condition = queries.Term(self.name, value)
        if self._inverted:
            return condition
        return queries.Bool(must_not=condition)

    @check_inversion
    def isin(self, values):
        """Condition checking whether multiple terms are in a field's value.

        Args:
            values (list): Sequence of values to check
        """
        return queries.Terms(self.name, values)

    def value_counts(self, n=10, normalize=False, missing=None, **es_kwargs):
        """Get the unique value counts for a field.

        The results are approximate if the number of unique terms is
        greater than the `n` parameter. The higher the requested `n` is,
        the more accurate the results will be, but also, the more expensive
        it will be to compute the final results.

        Args:
            n (int, optional): The number of value counts to return.
                Defaults to 10.
            normalize (bool, optional): If True then the object returned
                will contain the relative frequencies of the unique values.
                Defaults to False.
            missing (num, optional): How documents that are missing a
                value should be treated. Defaults to ignore.

        Returns:
            List[tuple]: Tuple consiting of (key, count) ordered by descending count.
                `OTHER` is included in counts if the `n` parameter is not
                enough to cover all counts.
        """
        results = self._simple_aggregation('terms',
                                           missing=missing,
                                           buckets=False,
                                           size=n,
                                           es_kwargs=es_kwargs)
        counts = [(i['key'], i['doc_count']) for i in results['buckets']]
        other_counts = results['sum_other_doc_count']
        error_bound = results['doc_count_error_upper_bound']
        if counts and error_bound > counts[-1][1]:
            msg = """
                The current `n` parameter may result in an approximation
                where a term is not included in the returned results ({})
            """.format(error_bound)
            warnings.warn(msg)
        if other_counts:
            counts.append(('OTHER', other_counts))
        if normalize:
            total = sum(i[1] for i in counts)
            counts = [(i[0], i[1] / float(total)) for i in counts]
        return counts

    def nunique(self, precision=3000, **es_kwargs):
        """Get the approximate count of distinct values.

        Args:
            precision (int, optional): Define a unique count below which
                counts are expected to be close to accurate. Trade memory
                for accuracy. Above this value, counts might become a bit
                more fuzzy. The maximum supported value is 40000, thresholds
                above this number will have the same effect as a threshold
                of 40000. Defaults to 3000.

        Returns:
            int: Distinct count
        """
        result = self._simple_aggregation('cardinality',
                                          precision_threshold=precision,
                                          es_kwargs=es_kwargs)
        return result['value']

    def _simple_aggregation(self, key, buckets=True, **params):
        """Execute simple aggregation query with no query hits."""
        if self.root._limit:
            warnings.warn("Limits are not applied in aggregations.")
        es_kwargs = params.pop('es_kwargs')
        params = {k: v for k, v in params.items() if v is not None}
        body = dict(self.root._body, **{
            'aggs': {
                key: {
                    key: dict({'field': self.name}, **params)
                }
            }
        })
        results = self.root.execute(body, size=0, **es_kwargs)
        results = results['aggregations'][key]
        if buckets:
            return results.get('buckets', results)
        return results


class AggregationMixin(object):
    """Adds aggregations to a field."""

    __metaclass__ = ABCMeta

    def average(self, **es_kwargs):
        """Get the average value for a field.

        Returns:
            float: Average
        """
        result = self._simple_aggregation('avg', es_kwargs=es_kwargs)
        return result['value']

    def max(self, **es_kwargs):
        """Get the max value for a field.

        Returns:
            float: Max
        """
        result = self._simple_aggregation('max', es_kwargs=es_kwargs)
        return result['value']

    def min(self, **es_kwargs):
        """Get the min value for a field.

        Returns:
            float: Max
        """
        result = self._simple_aggregation('min', es_kwargs=es_kwargs)
        return result['value']

    def percentiles(self, missing=None, precision=100, **es_kwargs):
        """Get percentiles for a filed.

        Percentiles are calculated over the range of [ 1, 5, 25, 50, 75, 95, 99 ]

        Args:
            missing (num, optional): How documents that are missing a
                value should be treated. Defaults to ignore.
            precision (int, optional): Defines the accuracy - memory tradeoff.
                Larger precision result in greater accuracy at the cost of
                greater memory usage and computation time. Defaults to 1000.

        Notes:
            - Accuracy is proportional to q(1-q). This means that extreme
                percentiles (e.g. 99%) are more accurate than less extreme
                percentiles, such as the median.
            - For small sets of values, percentiles are highly accurate
                (and potentially 100% accurate if the data is small enough).
            - As the quantity of values in a bucket grows, the algorithm
                begins to approximate the percentiles. It is effectively
                trading accuracy for memory savings.

        Returns:
            List[dict]: List of dicts with `key` as the perecentile range
                and `value` as the point where that percentile occurs.
        """
        result = self._simple_aggregation('percentiles',
                                          keyed=False,
                                          tdigest={'compression': precision},
                                          missing=missing,
                                          es_kwargs=es_kwargs)
        return result['values']

    def describe(self, extended=False, missing=None, **es_kwargs):
        """Get statistical details for a field.

        Args:
            extended (bool, optional): Increase the number of statistics returned.
                Defaults to False.
            missing (num, optional): How documents that are missing a
                value should be treated. Defaults to ignore.

        Returns:
            dict: Always returns `count`, `min`, `max`, `average`, `sum`.
                If `extended` is True then also returns `sum_of_squares`,
                `variance`, `std_deviation`, and `std_deviation_bounds`.

        """
        key = 'extended_stats' if extended else 'stats'
        return self._simple_aggregation(key, missing=missing, es_kwargs=es_kwargs)

    def histogram(self, interval=50, min_doc_count=1, missing=None, **es_kwargs):
        """Get a count of values for a field bucketed by interval.

        Args:
            interval (number, optional): Bin size for counts. Defaults to 50.
            min_doc_count (int, optional): Minimum number of documents for
                a bin to show up.
            missing (num, optional): How documents that are missing a
                value should be treated. Defaults to ignore.

        Returns:
            List[tuple]: Tuple consiting of (range, count).
        """
        def make_key(key):
            return '[{}, {})'.format(key, key+interval)
        results = self._simple_aggregation('histogram',
                                           interval=interval,
                                           min_doc_count=min_doc_count,
                                           missing=missing,
                                           es_kwargs=es_kwargs)
        return [(make_key(i['key']), i['doc_count']) for i in results]


class RangeMixin(object):
    """Adds comparator operators to a Field dtype."""

    __metaclass__ = ABCMeta

    @check_inversion
    def __lt__(self, value):
        return queries.Range(self.name).less_than(value)

    @check_inversion
    def __le__(self, value):
        return queries.Range(self.name).less_than_or_equal(value)

    @check_inversion
    def __gt__(self, value):
        return queries.Range(self.name).greater_than(value)

    @check_inversion
    def __ge__(self, value):
        return queries.Range(self.name).greater_than_or_equal(value)


class Numeric(Field, RangeMixin, AggregationMixin):
    """Numeric base for conditions and aggregations."""

    __metaclass__ = ABCMeta

    def percentile_ranks(self, values, missing=None, precision=100, **es_kwargs):
        """Get the percentage of observed values which are below a certain value.

        Args:
            values (List[int]): The values for which percentages are desired
            missing (num, optional): How documents that are missing a
                value should be treated. Defaults to ignore.
            precision (int, optional): Defines the accuracy - memory tradeoff.
                Larger precision result in greater accuracy at the cost of
                greater memory usage and computation time. Defaults to 1000.

        Notes:
            - Accuracy is proportional to q(1-q). This means that extreme
                percentiles (e.g. 99%) are more accurate than less extreme
                percentiles, such as the median.
            - For small sets of values, percentiles are highly accurate
                (and potentially 100% accurate if the data is small enough).
            - As the quantity of values in a bucket grows, the algorithm
                begins to approximate the percentiles. It is effectively
                trading accuracy for memory savings.

        Returns:
            List[dict]: List of dicts with `key` as the value and `value` as the
                percentage of items falling below that value.
        """
        result = self._simple_aggregation('percentile_ranks',
                                          keyed=False,
                                          values=values,
                                          tdigest={'compression': precision},
                                          missing=missing,
                                          es_kwargs=es_kwargs)
        return result['values']

    def sum(self, missing=None, **es_kwargs):
        """Get the sum of observed values.

        Args:
            missing (num, optional): How documents that are missing a
                value should be treated. Defaults to ignore.

        Returns:
            float: The sum
        """
        result = self._simple_aggregation('sum', missing=missing, es_kwargs=es_kwargs)
        return result['value']

    def median_absolute_deviation(self, missing=None, precision=1000, **es_kwargs):
        """Approximates the median absolute deviation for a field.

        Median absolute deviation is a measure of variability. It is a
        robust statistic, meaning that it is useful for describing data
        that may have outliers, or may not be normally distributed. For
        such data it can be more descriptive than standard deviation.

        https://en.wikipedia.org/wiki/Median_absolute_deviation

        Args:
            missing (num, optional): How documents that are missing a
                value should be treated. Defaults to ignore.
            precision (int, optional): Defines the accuracy - memory tradeoff.
                Larger precision result in greater accuracy at the cost of
                greater memory usage and computation time. Defaults to 1000.

        Returns:
            float: Median absolute deviation
        """
        result = self._simple_aggregation('median_absolute_deviation',
                                          compression=precision,
                                          missing=missing,
                                          es_kwargs=es_kwargs)
        return result['value']


class Integer(Numeric):
    """Integer field dtype."""

    dtype = 'integer'


class Float(Numeric):
    """Float field dtype."""

    dtype = 'float'


class Decimal(Numeric):
    """Decimal field dtype."""

    dtype = 'decimal'


class Boolean(Field):
    """Boolean field dtype."""

    dtype = 'boolean'


class String(Field):
    """String field dtype."""

    dtype = 'string'

    def match(self, term):
        """Condition checking whether term occurs in an analyzed field.

        Args:
            term (str): The term being matched
        """
        return queries.Match(self.name, term)

    @check_inversion
    def regexp(self, value):
        """Condition checking whether regular expression is to a field's value.

        Args:
            value (str): Regular expression
        """
        return queries.Regexp(self.name, value)

    @check_inversion
    def contains(self, value):
        """Condition checking whether string is contained in a field's value.

        This naive query is going to be significantly slower than using a
        regexp query or a match query with analyzed fields.

        Args:
            value (str): String that should exist
        """
        return queries.Wildcard(self.name, '*{}*'.format(value))

    @check_inversion
    def startswith(self, value):
        """Condition checking whether a field's value starts with a string.

        Args:
            value (str): String that a field's value should start with
        """
        return queries.Prefix(self.name, value)

    @check_inversion
    def endswith(self, value):
        """Condition checking whether a field's value ends with a string.

        This naive query is going to be significantly slower than using a
        regexp query or a match query with analyzed fields.

        Args:
            value (str): String that a field's value should end with
        """
        return queries.Wildcard(self.name, '*{}'.format(value))


class Date(Field, RangeMixin, AggregationMixin):
    """Date/datetime field dtype."""

    dtype = 'date'

    class Age(Field, RangeMixin):
        """Delta from datetime field dtype."""

        dtype = 'age'

        def age_to_dt(func):
            """Decorate a method to supply delta from today as date."""
            @wraps(func)
            def wrapper(obj, value):
                delta = datetime.today() - timedelta(value)
                return func(obj, delta)
            return wrapper

        __eq__ = age_to_dt(Field.__eq__)
        __ne__ = age_to_dt(Field.__ne__)
        __lt__ = age_to_dt(RangeMixin.__gt__)
        __le__ = age_to_dt(RangeMixin.__ge__)
        __gt__ = age_to_dt(RangeMixin.__lt__)
        __ge__ = age_to_dt(RangeMixin.__le__)

    @property
    def age(self):
        """Condition treating the value being compared as delta since today."""
        return self.Age(self.name, None)

    def _epoch_to_dt(func):
        @wraps(func)
        def wrapper(obj, *args, **kwargs):
            result = func(obj, *args, **kwargs)
            if result:
                return datetime.fromtimestamp(result / 1000)
        return wrapper

    average = _epoch_to_dt(AggregationMixin.average)
    max = _epoch_to_dt(AggregationMixin.max)
    min = _epoch_to_dt(AggregationMixin.min)

    def histogram(self, interval=50, min_doc_count=1):
        """Get a count of values for a field bucketed by interval.

        Args:
            interval (number, optional): Bin size for counts. Defaults to 50.
            min_doc_count (int, optional):
        """
        raise NotImplementedError("Bug Aaron to implement it.")


class Dummy(Field):
    """Dummy field for not implemented dtypes."""

    dtype = 'dummy'



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""Bamboo exceptions."""


class FieldConflictError(Exception):
    """Raise when a root field conflicts with a namespace."""

    msg = "Field name `{}` conflicts with a namespace in the mapping"

    def __init__(self, name):
        """Init FieldConflictError.

        Args:
            name (str): The name of the conflicting field
        """
        msg = self.msg.format(name)
        super(FieldConflictError, self).__init__(msg)


class MissingMappingError(Exception):
    """Raise when no mapping could be found for an index."""

    msg = "No mapping found for index `{}`"

    def __init__(self, index):
        """Init MissingMappingError.

        Args:
            index (str): The name of the index
        """
        msg = self.msg.format(index)
        super(MissingMappingError, self).__init__(msg)


class MissingQueryError(ValueError):
    """Raise when no query has been defined.

    Usually occurs when user is attempting to use an operator like and/or etc
    and no query currently is defined in the Dataframe object
    """

    msg = "No query found for operation"

    def __init__(self):
        """Init MissingQueryError."""
        super(MissingMappingError, self).__init__(self.msg)


class BadOperatorError(TypeError):
    """Raise when an inappropriate operator is used.

    Mostly to guard against someone using `x is True` instead of
    column-level `x == True`
    """

    msg = "Invalid operator: {}"

    def __init__(self, obj):
        """Init BadOperatorError.

        Args:
            obj (object): The object that caused this error to raise.
        """
        msg = self.msg.format(type(obj))
        super(BadOperatorError, self).__init__(msg)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""General utility functions."""
import functools
import warnings


def deprecated(func):
    """Decorate function as deprecated.

    It will result in a warning being emitted when the function is used.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        warnings.simplefilter('always', DeprecationWarning)  # turn off filter
        warnings.warn("{} is deprecated.".format(func.__name__),
                      category=DeprecationWarning,
                      stacklevel=2)
        warnings.simplefilter('default', DeprecationWarning)  # reset filter
        return func(*args, **kwargs)
    return wrapper


def dict_to_params(d):
    """Convert dictionary to string representation of parameters."""
    return ', ' .join('{}={}'.format(k, v) for k, v in d.items())



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""Provides a pandas-style interface for Elasticsearch.

Classes:
    DataFrame: Api for elasticsearch with pandas-style
        filtering operations
    ElasticDataFrame (deprecated): Api for elasticsearch with pandas-style
        filtering operations

Functions:
    boost: Boosts the weight of query by a value
    config: Accepts keyward arguments as configuration parameters

Exceptions:
    BadOperatorError: Raise when an inappropriate operator is used
    FieldConflictError: Raise when a root field conflicts with a namespace
    MissingMappingError: Raise when no mapping could be found for an index
    MissingQueryError: Raise when no query has been defined
"""
import pkg_resources

from .config import config
from .dataframe import DataFrame, ElasticDataFrame
from .exceptions import (BadOperatorError, FieldConflictError,
                         MissingMappingError, MissingQueryError)
from .queries import boost

__all__ = [
    'DataFrame',
    'ElasticDataFrame',

    'boost',
    'config',

    'BadOperatorError',
    'FieldConflictError',
    'MissingMappingError',
    'MissingQueryError'
]
__version__ = pkg_resources.get_distribution('elasticsearch-bamboo').version



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""Classes representing elasticsearch queries."""
from abc import ABCMeta, abstractproperty
from collections import defaultdict
from copy import deepcopy

from .utils import dict_to_params


def boost(query, value):
    """Boost the weight of the query by the value.

    Helps keep things clean when writing multiple querys by avoiding
    an extra pair of parenthesis.

    Instead of writing:
        df[(df.attr1 > 5).boost(2.0) & (df.attr2 < 2).boost(3.0)]
    You can write:
        df[boost(ds.attr1 > 5, 2.0) & boost(ds.attr2 < 2, 3.0)]
    """
    return query.boost(value)


class Query(object):
    """Base class for other query types.

    Attributes:
        field: Field the query is conditioned upon
        value: Value the condition is applying
        body: Query body formatted for elasticsearch
    """

    __metaclass__ = ABCMeta

    def __init__(self, field, value, boost=None):
        """Init Query.

        Args:
            field (str): Field the query is conditioned upon
            value (obj): Value the condition is applying
            boost (float, optional): Weight given to this query.
                Default None.
        """
        self.field = field
        self.value = value
        self._boost = boost

    @abstractproperty
    def key(self):
        """Get type of elasticsearch query object."""

    def __repr__(self):
        return '{}(field={}, value={}, boost={})'.format(
            type(self).__name__, self.field, self.value, self._boost
        )

    def __and__(self, other):
        if other is None:
            return self
        if isinstance(other, Bool):
            return Bool(must=self) & other
        return Bool(must=[self, other])
    __rand__ = __and__

    def __or__(self, other):
        if other is None:
            return self
        if isinstance(other, Bool):
            return Bool(should=self) | other
        return Bool(should=[self, other])
    __ror__ = __or__

    def __invert__(self):
        return Bool(must_not=self)

    def __call__(self):
        """Return the query body formatted for elasticsearch."""
        if self._boost is not None:
            return self._boosted_query
        return self._query

    @property
    def _query(self):
        return {
            self.key: {
                self.field: self.value
            }
        }

    @property
    def _boosted_query(self):
        return {
            self.key: {
                self.field: {
                    'value': self.value,
                    'boost': self._boost
                }
            }
        }

    def boost(self, value):
        """Boost the weight of the query by the value.

        Returns:
            Query: Query with boost applied.
        """
        new = deepcopy(self)
        new._boost = value
        return new


class Term(Query):
    """Find documents containing the exact term specified in the inverted index.

    The field is not analyzed.
    """

    key = 'term'


class Terms(Query):
    """Find documents containing any of the provided terms.

    The field is not analyzed.
    """

    key = 'terms'

    def __init__(self, field, value, boost=None):
        """Init Terms.

        Args:
            field (str): Field the query is conditioned upon
            value (list): List of terms to match
            boost (float, optional): Weight given to this query.
                Default None.
        """
        super(Terms, self).__init__(field, value, boost)

    @property
    def _boosted_query(self):
        return {
            self.key: {
                self.field: self.value,
                'boost': self._boost
            }
        }


class Regexp(Query):
    """Find documents that contain terms matching a regular expression."""

    key = 'regexp'


class Wildcard(Query):
    """Find documents that have fields matching a wildcard expression.

    The field is not analyzed.

    Supported wildcards are *, which matches any character sequence
    (including the empty one), and ?, which matches any single character.

    Note that this query can be slow, as it needs to iterate over many terms.
    In order to prevent extremely slow wildcard queries, a wildcard term
    should not start with one of the wildcards * or ?.
    """

    key = 'wildcard'


class Prefix(Query):
    """Find documents that contain a specific prefix in a provided field."""

    key = 'prefix'


class Match(Query):
    """Find documents that match a provided text, number, date or boolean value.

    The provided text is analyzed before matching.
    """

    key = 'match'

    @property
    def _boosted_query(self):
        return {
            self.key: {
                self.field: {
                    'query': self.value,
                    'boost': self._boost
                }
            }
        }


class Exists(Query):
    """Find documents that contain an indexed value for a field."""

    key = 'exists'

    def __init__(self, field, boost=None):
        """Init Exists.

        Args:
            field (str): Field the query is checking for existence
            boost (float, optional): Weight given to this query.
                Default None.
        """
        super(Exists, self).__init__('field', field, boost)

    @property
    def _boosted_query(self):
        return {
            self.key: {
                self.field: self.value,
                'boost': self._boost
            }
        }

    def __repr__(self):
        return 'Exists(field={}, boost={})'.format(self.value, self._boost)


class Range(Query):
    """Find documents that contain terms within a provided range."""

    key = 'range'
    _validation_msg = "At least one operation must be called."

    def __init__(self, field, boost=None):
        """Init Query.

        Args:
            field (str): Field the query is conditioned upon
            boost (float, optional): Weight given to this query.
                Default None.
        """
        self.field = field
        self._boost = boost
        self._operators = {}

    def greater_than(self, value):
        """Apply greater than operator to value.

        Returns:
            Range: Range query with greater than value applied.
        """
        new = deepcopy(self)
        new._operators['gt'] = value
        return new

    def greater_than_or_equal(self, value):
        """Apply greater than or equal operator to value.

        Returns:
            Range: Range query with greater than or equal value applied.
        """
        new = deepcopy(self)
        new._operators['gte'] = value
        return new

    def less_than(self, value):
        """Apply less than operator to value.

        Returns:
            Range: Range query with less than value applied.
        """
        new = deepcopy(self)
        new._operators['lt'] = value
        return new

    def less_than_or_equal(self, value):
        """Apply less than or equal operator to value.

        Returns:
            Range: Range query with less than or equal value applied.
        """
        new = deepcopy(self)
        new._operators['lte'] = value
        return new

    @property
    def _query(self):
        assert self._operators, self._validation_msg
        return {
            self.key: {
                self.field: self._operators
            }
        }

    @property
    def _boosted_query(self):
        assert self._operators, self._validation_msg
        return {
            self.key: {
                self.field: dict(self._operators, **{'boost': self._boost})
            }
        }

    def __repr__(self):
        return 'Range(field={}, boost={}, {})'.format(
            self.field,
            self._boost,
            dict_to_params(self._operators),
        )


class Bool(Query):
    """Find documents matching boolean combinations of other queries.

    Attributes:
        body: Query body formatted for elasticsearch
    """

    key = 'bool'
    _validation_msg = 'At least one initialization parameter must be supplied.'

    def __init__(self, must=None, filter=None, must_not=None, should=None, boost=None):
        """Init Bool.

        Args:
            must (list): The clause (query) must appear in matching documents
                and will contribute to the score.
            filter (list): The clause (query) must appear in matching
                documents. However unlike must the score of the query
                will be ignored. Scoring is ignored.
            should (list): The clause (query) should appear in the matching
                document. If the bool query has a must or filter clause
                then a document will match the bool query even if none of
                the should queries match. In this case these clauses are
                only used to influence the score. If the bool query has
                neither must or filter then at least one of the should
                queries must match a document for it to match the bool
                query.
            must_not (list): The clause (query) must not appear in the
                matching documents. Scoring is ignored.
            boost (float, optional): Weight given to this query.
                Default None.
        """
        self.params = dict(
            must=must or [],
            filter=filter or [],
            should=should or [],
            must_not=must_not or []
        )
        self._boost = boost
        self.__validate_params()

    def __repr__(self):
        return 'Bool(boost={}, {})'.format(self._boost, dict_to_params(self.params))

    def __validate_params(self):
        """Ensure parameters are in correct format."""
        assert any(self.params.values()), self._validation_msg
        for key, params in self.params.items():
            if isinstance(params, Query):
                self.params[key] = [params]
            elif not isinstance(params, list):
                self.params[key] = list(params)

    @property
    def must(self):
        """Get the must conditions."""
        return self.params['must']

    @property
    def must_not(self):
        """Get the must_not conditions."""
        return self.params['must_not']

    @property
    def should(self):
        """Get the should conditions."""
        return self.params['should']

    @property
    def filter(self):
        """Get the filter conditions."""
        return self.params['filter']

    def __add__(self, other):
        """Merge the parameters of two queries into one."""
        if other is None:
            return self
        if not isinstance(other, Bool):
            other = Bool(must=other)

        params = defaultdict(list, deepcopy(self.params))
        for k, v in other.params.items():
            params[k].extend(v)
        return Bool(**params)
    __radd__ = __add__

    def __and__(self, other):
        if not isinstance(other, Bool):
            other = Bool(must=other)
        return Bool(must=self.explode('must') + other.explode('must'))

    def __or__(self, other):
        if not isinstance(other, Bool):
            other = Bool(should=other)
        return Bool(should=self.explode('should') + other.explode('should'))

    def explode(self, destination):
        """Explode all parameters into individual Bool queries.

        This is used for boolean logic operations between two Bool queries.
        Parameters named by `destination` are left as is.

        Example:
            >>> self
            Bool(must=[Term('attr', 1), should=[Term('attr', 2)], must_not=[Term('attr', 3)])
            >>> self.compress('should')
            [Term('attr', 2),
             Bool(must=[Term('attr', 1)],
             Bool(must_not=[Term('attr', 3)]

        Args:
            destination (str): The parameter that all other paremeters
                are converted to

        Returns:
            List[Bool]: List of Bool queries which were previously the
                objects individual parameters
        """
        params = self.filtered_params
        compressed = params.pop(destination, [])
        for name, clauses in params.items():
            compressed.append(Bool(**{name: clauses}))
        return compressed

    def __invert__(self):
        return Bool(
            must=[self.__negate(i) for i in self.should] + self.must_not,
            should=[self.__negate(i) for i in self.must + self.filter]
        )

    def __negate(self, query):
        """Convert `must` queries to `must_not`."""
        if isinstance(query, Bool):
            return ~query
        return Bool(must_not=query)

    @property
    def _query(self):
        # # if must is only parameter and contains only one condition
        # # then convert that one condition to a stand-alone query
        if len(self.filtered_params) == 1 == len(self.must):
            return self.must[0]()
        return {self.key: self._finalize_params()}

    @property
    def _boosted_query(self):
        # if must is only parameter and contains only one condition
        # then convert that one condition to a stand-alone query
        if len(self.filtered_params) == 1 == len(self.must):
            return self.must[0].boost(self._boost)()
        return {self.key: dict(self._finalize_params(), **{'boost': self._boost})}

    @property
    def filtered_params(self):
        """Get only the parameters populated with conditions."""
        return {k: v for k, v in self.params.items() if v}

    def _finalize_params(self):
        return {
            key: [param() for param in params]
            for key, params in self.params.items()
            if params
        }


class Script(Query):
    """Find documents matching a painless scripts as a query."""

    key = 'script'

    def __init__(self, source, boost=None):
        """Init Exists.

        Args:
            source (str): Painless script
            boost (float, optional): Weight given to this query.
                Default None.
        """
        self.source = source
        self._boost = boost

    def __repr__(self):
        return "Script(source='{}', boost={})".format(self.source, self._boost)

    @property
    def _query(self):
        return {
            self.key: {
                self.key: {
                    'source': self.source,
                    'lang': 'painless'
                }
            }
        }

    @property
    def _boosted_query(self):
        return {
            self.key: {
                self.key: {
                    'source': self.source,
                    'lang': 'painless'
                },
                'boost': self._boost
            }
        }



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""Install bamboo package."""
from setuptools import setup

VERSION = '0.3.2'

setup(
    name='elasticsearch-bamboo',
    packages=['bamboo'],
    version=VERSION,
    license='Mozilla Public License 2.0 (MPL 2.0)',
    description='DataFrame interface for ElasticSearch',
    author='Aaron Mangum',
    author_email='aaron.mangum@juvo.com',
    url='https://github.com/juvoinc/bamboo',
    download_url='https://github.com/juvoinc/bamboo/archive/{}.tar.gz'.format(VERSION),
    keywords=['elasticsearch', 'dataframe', 'pandas'],
    python_requires='>=2.7',
    install_requires=[
        'elasticsearch>=6.0.0',
        'six'  # temp py2 support
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Topic :: Software Development',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
    ],
    extras_require=dict(
        dev=[
            'coverage',
            'flake8',
            'flake8-docstrings',
            'pytest',
            'tox'
        ],
        test=[
            'pandas',
            'pytest',
        ],
        pandas=['pandas']
    )
)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
"""Pandas-style framework for interacting with elasticsearch."""
from copy import deepcopy

from six import string_types

from elasticsearch.helpers import scan

from .config import config
from .exceptions import BadOperatorError, MissingQueryError
from .orm import OrmMixin
from .queries import Bool, Query, Script


class DataFrame(OrmMixin):
    """Api for elasticsearch with pandas-style filtering operations.

    Attributes:
        index: The elasticsearch index name
        config: Configuration object for elasticsearch
    """

    config = config

    def __init__(self, index, frozen=True, config=None):
        """Init DataFrame.

        Args:
            index (str): The name of the index to use
            frozen (bool, optional): Whether the dataframe is immutable.
                Defaults to True.
            config (config.Config.optional): Configuration object. If none
                provided then uses mutable config.config in global namespace.
                Defaults to None.
        """
        if not frozen:
            raise NotImplementedError
        self.index = index
        self.config = config or self.config
        self._query = None
        self._limit = None
        self._load_orm()

    @property
    def _es(self):
        return self.config.connection

    def __repr__(self):
        df = self.__pandas_100()
        if df is not None:
            return repr(df)
        return '{}({})'.format(type(self).__name__, self.index)

    def _repr_html_(self):
        """Return an html representation for a particular DataFrame.

        Mainly for IPython notebook.
        """
        df = self.__pandas_100()
        if df is not None:
            return df._repr_html_()
        return None

    def __pandas_100(self):
        frame = self if self._limit else self.limit(100)
        try:
            return frame.to_pandas()
        except ImportError:
            pass

    def __getitem__(self, item):
        if isinstance(item, string_types):
            return getattr(self, item)
        if isinstance(item, Query):
            new = deepcopy(self)
            new._query &= item
            return new
        raise BadOperatorError(item)

    def __and__(self, other):
        new = deepcopy(self)
        new._query &= other._query
        return new

    __add__ = __and__

    def __or__(self, other):
        new = deepcopy(self)
        new._query |= other._query
        return new

    def __invert__(self):
        new = deepcopy(self)
        if new._query is None:
            raise MissingQueryError
        new._query = ~new._query
        return new

    def get(self, id, fields=None):
        """Return the document represented by an ID.

        Args:
            id (str): ID for a document
            fields (list): The field names that should be returned from source.
                If None then returns all. Defaults to None.
                Note: Namespaced fields should be referenced using `.` syntax
                    (`<namespace>.<field>`). They will be returned using the
                    same syntax.

        Returns:
            dict: Document source for a given ID
        """
        doc = self._es.get(index=self.index,
                           id=id,
                           doc_type='doc',
                           _source=fields)
        return doc['_source']

    __call__ = get

    @classmethod
    def list_indices(cls):
        """List all the indices in the elasticsearch cluster."""
        return cls.config.connection.indices.get('*').keys()

    def query(self, source):
        """Set condition according to painless script.

        Args:
            source (str): Parseable expression string

        Returns:
            query.Script: Script query object.
        """
        return Script(source)

    def limit(self, n):
        """Limit the number of results returned by elasticsearch.

        Args:
            n (int): The number of results to return

        Returns:
            DataFrame: DataFrame with limit applied.
        """
        new = deepcopy(self)
        new._limit = n
        return new

    def filter(self, *conditions):
        """Each condition must appear in matching documents.

        This does not contribute to scoring.

        Args:
            *conditions (query): Query conditions that will be matched

        Returns:
            DataFrame: DataFrame with filter applied.
        """
        new = deepcopy(self)
        if any(isinstance(i, DataFrame) for i in conditions):
            raise TypeError(DataFrame)
        new._query += Bool(filter=conditions)
        return new

    @property
    def _body(self):
        """Raw query sent to elasticsearch."""
        if not self._query:
            return {'query': {'match_all': {}}}
        return {'query': self._query()}

    def execute(self, body, size=None, fields=None,
                preserve_order=False, **es_kwargs):
        """Execute elasticsearch query.

        Args:
            body (dict): Query body in json format
            size (int, optional): The number of results to return.
                 If None then returns a generator scan of the entire index.
                 Defaults to None.
            fields (List[str], optional): The fields to include in the document
                _source in the return results. Defaults to None.
            preserve_order (bool, optional): Whether to return the results
                in sorted order. Only applies where size is None, otherwise
                always returns results in sorted order. Defaults to False.

        Returns:
            List[dict]: The raw elasticsearch results. Returns a generator
                if `size` is None.
        """
        if size is None:
            return scan(
                client=self._es,
                index=self.index,
                query=body,
                preserve_order=preserve_order,
                _source=fields,
                **es_kwargs
            )
        return self._es.search(
            index=self.index,
            body=body,
            size=size,
            _source=fields,
            **es_kwargs
        )

    def collect(self, fields=None, limit=None, preserve_order=False,
                include_score=False, include_id=False, **es_kwargs):
        """Collect documents according to query conditions.

        Args:
            fields (list): The field names that should be returned from source.
                If None then returns all. Defaults to None.
                Note: Namespaced fields should be referenced using `.` syntax
                    (`<namespace>.<field>`). They will be returned using the
                    same syntax.
            limit (int, optional): The number of results to return. If a limit
                is set then the documents will always be in sorted order.
                Defaults to None.
            preserve_order (bool, optional): Whether to keep the documents
                in sorted order. This will only be applied where no limit has
                been set on the number of documents. Defaults to False.
                Note: This may be expensive on large queries.
            include_score (bool, optional): Whether to include the document score in
                the results. This can be computationally expensive. Default False.
            include_id (bool, optional): Whether to include the document id
                in the results. Default False.
            **es_kwargs (dict, optional): Additional arguments to pass to elasticsearch.

        Returns:
            generator: The response from elasticsearch
        """
        size = limit or self._limit
        body = dict(self._body, **{'track_scores': include_score})
        results = self.execute(body, size, fields, preserve_order, **es_kwargs)
        return self.__hits(results, include_score, include_id)

    def count(self):
        """Return the count of documents that match.

        Returns:
            int: The number of documents
        """
        results = self._es.count(index=self.index, body=self._body)
        return results['count']

    def take(self, n, fields=None):
        """Collect documents according to query conditions.

        Args:
            n (int, optional): The number of results to return. Defaults to None.
            fields (list): The field names that should be returned from source.
                If None then returns all. Defaults to None.
                Note: Namespaced fields should be referenced using `.` syntax
                    (`<namespace>.<field>`). They will be returned using the
                    same syntax.

        Returns:
            list: The response from elasticsearch
        """
        return list(self.collect(fields=fields, limit=n))

    def __hits(self, results, include_score, include_id):
        """Format the raw elasticsearch results to return just source."""
        results = results['hits']['hits'] if isinstance(results, dict) else results
        for hit in results:
            result = hit['_source']
            if include_score:
                result['_score'] = hit.pop('_score')
            if include_id:
                result['_id'] = hit.pop('_id')
            yield result

    def to_pandas(self, fields=None):
        """Collect documents according to query conditions as a pandas DataFrame.

        Args:
            fields (list): The field names that should be returned from source.
                If None then returns all. Defaults to None.
                Note: Namespaced fields should be referenced using `.` syntax
                    (`<namespace>.<field>`). They will be returned using the
                    same syntax.

        Returns:
            pd.DataFrame: The response from elasticsearch
        """
        try:
            import pandas as pd
        except ImportError:
            raise ImportError('Install pandas for pandas support.')
        else:
            data = self.collect(fields, preserve_order=False)
            data = (self.__nested_to_dot(i) for i in data)
            return pd.DataFrame(data)

    @classmethod
    def __nested_to_dot(cls, hit, namespace=''):
        """Recursively flattens nested objects to use dot-notation."""
        d = {}
        for k, v in hit.items():
            key = '{}.{}'.format(namespace, k) if namespace else k
            if isinstance(v, dict):
                d.update(cls.__nested_to_dot(v, key))
            else:
                d[key] = v
        return d


class ElasticDataFrame(DataFrame):
    def __new__(cls, *args, **kwargs):
        from warnings import warn
        warn("{} is deprecated. Use DataFrame instead.".format(cls.__name__),
             category=DeprecationWarning,
             stacklevel=2)
        return DataFrame(*args, **kwargs)
